@@ -46,6 +46,7 @@ public:
 
     SurfaceInterface *surface() const;
     wl_shm_buffer *shmBuffer();
+    wl_resource *resource() const;
 
     /**
      * Creates a QImage for the shared memory buffer.
@@ -69,8 +70,34 @@ public:
      **/
     QImage data();
 
+    /**
+     * Returns the size of this BufferInterface.
+     * Note: only for shared memory buffers (shmBuffer) the size can be derived,
+     * for other buffers it might be possible to derive the size if an EGL display
+     * is set on Display otherwise the user of the BufferInterface has to use setSize to
+     * provide the proper size.
+     * @see setSize
+     * @see Display::setEglDisplay
+     * @since 5.3
+     **/
+    QSize size() const;
+    /**
+     * Sets the @p size for non shared memory buffers.
+     * @see size
+     * @see sizeChanged
+     * @since 5.3
+     **/
+    void setSize(const QSize &size);
+
+    static BufferInterface *get(wl_resource *r);
+
 Q_SIGNALS:
     void aboutToBeDestroyed(KWayland::Server::BufferInterface*);
+    /**
+     * Emitted when the size of the Buffer changes.
+     * @since 5.3
+     **/
+    void sizeChanged();
 
 private:
     friend class SurfaceInterface;
