@@ -39,8 +39,13 @@ public:
     SeatInterface *seat;
     SurfaceInterface *focusedSurface = nullptr;
     QMetaObject::Connection destroyConnection;
+    Cursor *cursor = nullptr;
 
 private:
+    PointerInterface *q_func() {
+        return reinterpret_cast<PointerInterface *>(q);
+    }
+    void setCursor(quint32 serial, SurfaceInterface *surface, const QPoint &hotspot);
     // interface
     static void setCursorCallback(wl_client *client, wl_resource *resource, uint32_t serial,
                                   wl_resource *surface, int32_t hotspot_x, int32_t hotspot_y);
@@ -50,12 +55,41 @@ private:
     static const struct wl_pointer_interface s_interface;
 };
 
+class Cursor::Private
+{
+public:
+    Private(Cursor *q, PointerInterface *pointer);
+    PointerInterface *pointer;
+    quint32 enteredSerial = 0;
+    QPoint hotspot;
+    QPointer<SurfaceInterface> surface;
+
+    void update(const QPointer<SurfaceInterface> &surface, quint32 serial, const QPoint &hotspot);
+
+private:
+    Cursor *q;
+};
+
 PointerInterface::Private::Private(SeatInterface *parent, wl_resource *parentResource, PointerInterface *q)
     : Resource::Private(q, parent, parentResource, &wl_pointer_interface, &s_interface)
     , seat(parent)
 {
 }
 
+void PointerInterface::Private::setCursor(quint32 serial, SurfaceInterface *surface, const QPoint &hotspot)
+{
+    if (!cursor) {
+        Q_Q(PointerInterface);
+        cursor = new Cursor(q);
+        cursor->d->enteredSerial = serial;
+        cursor->d->hotspot = hotspot;
+        cursor->d->surface = QPointer<SurfaceInterface>(surface);
+        QObject::connect(cursor, &Cursor::changed, q, &PointerInterface::cursorChanged);
+        emit q->cursorChanged();
+    } else {
+        cursor->d->update(QPointer<SurfaceInterface>(surface), serial, hotspot);
+    }
+}
 
 const struct wl_pointer_interface PointerInterface::Private::s_interface = {
     setCursorCallback,
@@ -129,13 +163,9 @@ void PointerInterface::axis(Qt::Orientation orientation, quint32 delta)
 void PointerInterface::Private::setCursorCallback(wl_client *client, wl_resource *resource, uint32_t serial,
                                          wl_resource *surface, int32_t hotspot_x, int32_t hotspot_y)
 {
-    Q_UNUSED(client)
-    Q_UNUSED(resource)
-    Q_UNUSED(serial)
-    Q_UNUSED(surface)
-    Q_UNUSED(hotspot_x)
-    Q_UNUSED(hotspot_y)
-    // TODO: implement
+    auto p = cast<Private>(resource);
+    Q_ASSERT(p->client->client() == client);
+    p->setCursor(serial, SurfaceInterface::get(surface), QPoint(hotspot_x, hotspot_y));
 }
 
 void PointerInterface::Private::releaseCallback(wl_client *client, wl_resource *resource)
@@ -144,10 +174,79 @@ void PointerInterface::Private::releaseCallback(wl_client *client, wl_resource *
     unbind(resource);
 }
 
+Cursor *PointerInterface::cursor() const
+{
+    Q_D();
+    return d->cursor;
+}
+
 PointerInterface::Private *PointerInterface::d_func() const
 {
     return reinterpret_cast<Private*>(d.data());
 }
 
+Cursor::Private::Private(Cursor *q, PointerInterface *pointer)
+    : pointer(pointer)
+    , q(q)
+{
+}
+
+void Cursor::Private::update(const QPointer< SurfaceInterface > &s, quint32 serial, const QPoint &p)
+{
+    bool emitChanged = false;
+    if (enteredSerial != serial) {
+        enteredSerial = serial;
+        emitChanged = true;
+        emit q->enteredSerialChanged();
+    }
+    if (hotspot != p) {
+        hotspot = p;
+        emitChanged = true;
+        emit q->hotspotChanged();
+    }
+    if (surface != s) {
+        if (!surface.isNull()) {
+            QObject::disconnect(surface.data(), &SurfaceInterface::damaged, q, &Cursor::changed);
+        }
+        surface = s;
+        if (!surface.isNull()) {
+            QObject::connect(surface.data(), &SurfaceInterface::damaged, q, &Cursor::changed);
+        }
+        emitChanged = true;
+        emit q->surfaceChanged();
+    }
+    if (emitChanged) {
+        emit q->changed();
+    }
+}
+
+Cursor::Cursor(PointerInterface *parent)
+    : QObject(parent)
+    , d(new Private(this, parent))
+{
+}
+
+Cursor::~Cursor() = default;
+
+quint32 Cursor::enteredSerial() const
+{
+    return d->enteredSerial;
+}
+
+QPoint Cursor::hotspot() const
+{
+    return d->hotspot;
+}
+
+PointerInterface *Cursor::pointer() const
+{
+    return d->pointer;
+}
+
+QPointer< SurfaceInterface > Cursor::surface() const
+{
+    return d->surface;
+}
+
 }
 }
